from abc import abstractmethod
from typing import Callable

from DataStructure.Heap.HeapNode import HeapNode


class Heap:

    __array: list
    __count: int
    comparator: Callable[[object, object], int]

    def __init__(self, N: int, comparator: Callable[[object, object], int]):
        self.comparator = comparator
        self.__count = 0
        self.__array = []
        for i in range(N):
            self.__array.append(None)

    @abstractmethod
    def compare(self, data1: object, data2: object) -> int:
        pass

    def isEmpty(self) -> bool:
        return self.__count == 0

    def swapNode(self, index1: int, index2: int):
        tmp = self.__array[index1]
        self.__array[index1] = self.__array[index2]
        self.__array[index2] = tmp

    def percolateDown(self, no: int):
        left = 2 * no + 1
        right = 2 * no + 2
        while (left < self.__count and self.compare(self.__array[no].getData(), self.__array[left].getData()) < 0) or \
                (right < self.__count and self.compare(self.__array[no].getData(), self.__array[right].getData()) < 0):
            if right >= self.__count or self.compare(self.__array[left].getData(), self.__array[right].getData()) > 0:
                self.swapNode(no, left)
                no = left
            else:
                self.swapNode(no, right)
                no = right
            left = 2 * no + 1
            right = 2 * no + 2

    def percolateUp(self, no: int):
        parent = (no - 1) // 2
        while parent >= 0 and self.compare(self.__array[parent].getData(), self.__array[no].getData()) < 0:
            self.swapNode(parent, no)
            no = parent
            parent = (no - 1) // 2

    def delete(self) -> object:
        tmp = self.__array[0]
        self.__array[0] = self.__array[self.__count - 1]
        self.percolateDown(0)
        self.__count = self.__count - 1
        return tmp.getData()

    def insert(self, data: object):
        self.__count = self.__count + 1
        self.__array[self.__count - 1] = HeapNode(data)
        self.percolateUp(self.__count - 1)



from typing import Callable

from DataStructure.Stack import Stack
from DataStructure.Tree.AvlTreeNode import AvlTreeNode
from DataStructure.Tree.Tree import Tree


class AvlTree(Tree):

    def __init__(self, comparator: Callable[[object, object], int]):
        super().__init__(comparator)

    def height(self, d: AvlTreeNode) -> int:
        if d is None:
            return 0
        else:
            return d.height

    def rotateLeft(self, k2: AvlTreeNode) -> AvlTreeNode:
        k1 = k2.left
        k2.left = k1.right
        k1.right = k2
        k2.height = max(self.height(k2.left), self.height(k2.right)) + 1
        k1.height = max(self.height(k1.left), k1.right.height) + 1
        return k1

    def rotateRight(self, k1: AvlTreeNode) -> AvlTreeNode:
        k2 = k1.right
        k1.right = k2.left
        k2.left = k1
        k2.height = max(k2.left.height, self.height(k2.right)) + 1
        k1.height = max(self.height(k1.left), self.height(k1.right)) + 1
        return k2

    def doubleRotateLeft(self, k3: AvlTreeNode):
        k3.left = self.rotateRight(k3.left)
        return self.rotateLeft(k3)

    def doubleRotateRight(self, k1: AvlTreeNode):
        k1.right = self.rotateLeft(k1.right)
        return self.rotateRight(k1)

    def insert(self, node: AvlTreeNode):
        LEFT = 1
        RIGHT = 2
        y = None
        x = self.root
        dir1 = 0
        dir2 = 0
        c = Stack()
        while x is not None:
            y = x
            c.push(y)
            dir1 = dir2
            if self.comparator(node.data, x.data) < 0:
                x = x.left
                dir2 = LEFT
            else:
                x = x.right
                dir2 = RIGHT
        self.insertChild(y, node)
        while not c.isEmpty():
            x = c.pop()
            x.height = max(self.height(x.left), self.height(x.right)) + 1
            if abs(self.height(x.left) - self.height(x.right)) == 2:
                if dir1 == LEFT:
                    if dir2 == LEFT:
                        t = self.rotateLeft(x)
                    else:
                        t = self.doubleRotateLeft(x)
                else:
                    if dir2 == LEFT:
                        t = self.doubleRotateRight(x)
                    else:
                        t = self.rotateRight(x)
                y = c.pop()
                self.insertChild(y, t)
                break

    def insertData(self, data: object):
        self.insert(AvlTreeNode(data))



class Stack:

    __stack: list

    def __init__(self):
        self.__stack = []

    def push(self, item: object):
        self.__stack.append(item)

    def pop(self) -> object:
        if len(self.__stack) > 0:
            return self.__stack.pop()
        else:
            return None

    def isEmpty(self) -> bool:
        return len(self.__stack) == 0



from typing import Callable

from DataStructure.Tree.TreeNode import TreeNode


class Tree:

    root: TreeNode = None
    comparator: Callable[[object, object], int]

    def __init__(self, comparator: Callable[[object, object], int]):
        self.comparator = comparator

    def search(self, value: object) -> TreeNode:
        d = self.root
        while d is not None:
            if self.comparator(d.data, value) == 0:
                return d
            else:
                if self.comparator(d.data, value) > 0:
                    d = d.left
                else:
                    d = d.right
        return None

    def insertChild(self, parent: TreeNode, child: TreeNode):
        if parent is None:
            self.root = child
        else:
            if self.comparator(child.data, parent.data) < 0:
                parent.left = child
            else:
                parent.right = child

    def insert(self, node: TreeNode):
        y = None
        x = self.root
        while x is not None:
            y = x
            if self.comparator(node.data, x.data) < 0:
                x = x.left
            else:
                x = x.right
        self.insertChild(y, node)

    def insertData(self, data: object):
        self.insert(TreeNode(data))



from __future__ import annotations

from typing import Callable


class BTreeNode:

    K: list
    m: int
    d: int
    leaf: bool
    children: [BTreeNode]

    def __init__(self, d: int, firstChild: BTreeNode = None, secondChild: BTreeNode = None, newK: object = None):
        self.d = d
        self.K = []
        self.children = []
        if firstChild is None:
            self.m = 0
            self.leaf = True
        else:
            self.leaf = False
            self.m = 1
            self.children.append(firstChild)
            self.children.append(secondChild)
            self.K.append(newK)

    def position(self, value: object, comparator: Callable[[object, object], int]) -> int:
        if self.m == 0:
            return 0
        if comparator(value, self.K[self.m - 1]) > 0:
            return self.m
        else:
            for i in range(self.m):
                if comparator(value, self.K[i]) <= 0:
                    return i
        return -1

    def insertIntoK(self, index: int, insertedK: object):
        if index < len(self.K):
            for i in range(self.m, index, -1):
                self.K[i] = self.K[i - 1]
            self.K[index] = insertedK
        else:
            self.K.append(insertedK)

    def moveHalfOfTheKToNewNode(self, newNode: BTreeNode):
        for i in range(self.d):
            newNode.K.append(self.K[i + self.d + 1])
        newNode.m = self.d

    def moveHalfOfTheChildrenToNewNode(self, newNode: BTreeNode):
        for i in range(self.d):
            newNode.children.append(self.children[i + self.d + 1])

    def moveHalfOfTheElementsToNewNode(self, newNode: BTreeNode):
        self.moveHalfOfTheKToNewNode(newNode)
        self.moveHalfOfTheChildrenToNewNode(newNode)

    def insertNode(self, value: object, comparator: Callable[[object, object], int], isRoot: bool) -> BTreeNode:
        child = self.position(value, comparator)
        if not self.children[child].leaf:
            s = self.children[child].insertNode(value, comparator, False)
        else:
            s = self.children[child].insertLeaf(value, comparator)
        if s is None:
            return None
        self.insertIntoK(child, self.children[child].K[self.d])
        if self.m < 2 * self.d:
            self.children.append(s)
            self.m = self.m + 1
            return None
        else:
            newNode = BTreeNode(self.d)
            newNode.leaf = False
            self.moveHalfOfTheElementsToNewNode(newNode)
            newNode.children.append(s)
            self.m = self.d
            if isRoot:
                a = BTreeNode(self.d, self, newNode, self.K[self.d])
                return a
            else:
                return newNode

    def insertLeaf(self, value: object, comparator: Callable[[object, object], int]) -> BTreeNode:
        child = self.position(value, comparator)
        self.insertIntoK(child, value)
        if self.m < 2 * self.d:
            self.m = self.m + 1
            return None
        else:
            newNode = BTreeNode(self.d)
            self.moveHalfOfTheKToNewNode(newNode)
            self.m = self.d
            return newNode



from typing import Callable

from DataStructure.Tree.BTreeNode import BTreeNode


class BTree:

    root: BTreeNode = None
    comparator: Callable[[object, object], int]
    d: int

    def __init__(self, d: int, comparator: Callable[[object, object], int]):
        self.comparator = comparator
        self.d = d

    def search(self, value: object) -> BTreeNode:
        b = self.root
        while not b.leaf:
            child = b.position(value, self.comparator)
            if child < b.m and b.K[child] == value:
                return b
            b = b.children[child]
        child = b.position(value, self.comparator)
        if child < b.m and b.K[child] == value:
            return b
        return None

    def insertData(self, data: object):
        if self.root is None:
            self.root = BTreeNode(self.d)
        if self.root.leaf:
            s = self.root.insertLeaf(data, self.comparator)
            if s is not None:
                tmp = self.root
                self.root = BTreeNode(self.d, tmp, s, tmp.K[self.d])
        else:
            s = self.root.insertNode(data, self.comparator, True)
            if s is not None:
                self.root = s



from typing import Callable

from DataStructure.Heap.Heap import Heap


class MaxHeap(Heap):

    def __init__(self, N: int, comparator: Callable[[object, object], int]):
        super().__init__(N, comparator)

    def compare(self, data1: object, data2: object) -> int:
        return self.comparator(data1, data2)
